unit Bifurcation;

interface

uses
  Windows, Messages, SysUtils, Classes, ASIConst, RPNCompiler;

type
  TBifurcation = class(TComponent)
  private
    { Private declarations }
    FSeed: Extended;
    procedure EvalBif(X: Extended; var N: Integer; P: PExtendedArray0);
    function EvalLCE(X: Extended): Extended;
    function Eval(X: Extended; L: Extended): Extended;
  protected
    { Protected declarations }
    RPN: TRPNCompiler;
    RPN_Der: TRPNCompiler;
    Mutex: TMultiReadExclusiveWriteSynchronizer;
    function GetBiforcazione: TMultiFunzione;
    function GetLyapunov: TFunzione;
    function GetIterazione(L: Extended): TFunzione;
    procedure SetSeed(Value: Extended);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Logistic: TRPNCompiler read RPN;
    property Derivata: TRPNCompiler read RPN_Der;

    property Biforcazione: TMultiFunzione read GetBiforcazione;
    property Iterazione[L: Extended]: TFunzione read GetIterazione;
    property Lyapunov: TFunzione read GetLyapunov;
    property Seed: Extended read FSeed write SetSeed;

  published
    { Published declarations }
  end;

implementation

uses Math;

type
  TBiforcazione = class(TMultiFunzione)
    Bifurcation: TBifurcation;
    constructor Create(ABifurcation: TBifurcation);
    procedure Valuta(X: Extended; var N: Integer; P: PExtendedArray0); override;
  end;

type
  TLyapunov = class(TFunzione)
    Bifurcation: TBifurcation;
    constructor Create(ABifurcation: TBifurcation);
    function Valuta(X: Extended): Extended; override;
  end;

type
  TIterazione = class(TFunzione)
    Bifurcation: TBifurcation;
    L: Extended;
    constructor Create(ABifurcation: TBifurcation; AL: Extended);
    function Valuta(X: Extended): Extended; override;
  end;

constructor TBiforcazione.Create(ABifurcation: TBifurcation);
begin
  inherited Create;
  Bifurcation := ABifurcation;
end;

procedure TBiforcazione.Valuta(X: Extended; var N: Integer; P: PExtendedArray0);
begin
  Bifurcation.EvalBif(X, N, P);
end;

constructor TLyapunov.Create(ABifurcation: TBifurcation);
begin
  inherited Create;
  Bifurcation := ABifurcation;
end;

function TLyapunov.Valuta(X: Extended): Extended;
begin
  Result := Bifurcation.EvalLCE(X);
end;

constructor TIterazione.Create(ABifurcation: TBifurcation; AL: Extended);
begin
  inherited Create;
  L := AL;
  Bifurcation := ABifurcation;
end;

function TIterazione.Valuta(X: Extended): Extended;
begin
  Result := Bifurcation.Eval(X, L);
end;

constructor TBifurcation.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  RPN := TRPNCompiler.Create(nil);
  RPN_Der := TRPNCompiler.Create(nil);
  Mutex := TMultiReadExclusiveWriteSynchronizer.Create;
  FSeed := 0.321;
end;

destructor TBifurcation.Destroy;
begin
  RPN.Free;
  RPN_Der.Free;
  Mutex.Free;
  inherited Destroy;
end;

function TBifurcation.Eval(X: Extended; L: Extended): Extended;
begin
  Result := RPN.Eval([L, X]);
end;

procedure TBifurcation.EvalBif(X: Extended; var N: Integer; P: PExtendedArray0);
var
  I, F, T: Integer;
  Y: Extended;
  Ok: Boolean;
begin
  Y := FSeed;

  T := N;
  F := 0;

  for I := 0 to 10000 do
    Y := RPN.Eval([X, Y]);

  repeat
    N := 0;
    repeat
      P[N] := Y;
      Y := RPN.Eval([X, Y]);
      Inc(N);
      Inc(F);
      Ok := SameValue(Y, P[0], Y * 0.0001);
    until Ok or (N = T);
  until Ok or (F > 500);

  if IsNaN(P[0]) or IsInfinite(P[0]) then
    raise Exception.Create('No period!');

end;

function TBifurcation.EvalLCE(X: Extended): Extended;
const
  EPS = 1.0e-4;
  MAXPERIOD = 129;
var
  Der: Extended;
  I: Integer;
  Sum1, Sum2, Sum3, Tmp: Extended;
  P: PExtendedArray0;
  N: Integer;
begin
  GetMem(P, SizeOf(Extended) * MAXPERIOD);
  try

    N := MAXPERIOD;

    EvalBif(X, N, P);
    Sum2 := 0;
    Sum3 := 0;
    for I := 0 to N - 1 do begin
      Der := RPN_Der.Eval([X, P[I]]);
      if Der = 0 then
        raise Exception.Create('No LCE!');

      Sum1 := Sum2;
      Sum2 := Sum3;
      Sum3 := Sum3 + Ln(Abs(Der));
    end;

    Tmp := Sum3 - Sum2;
    Result := Sum3 - (Tmp * Tmp) / (Sum3 - 2 * Sum2 + Sum1);
    Result := Result / N;

    if IsNaN(Result) or IsInfinite(Result) then
      raise Exception.Create('No LCE!');

  finally
    FreeMem(P);
  end;

end;

function TBifurcation.GetBiforcazione: TMultiFunzione;
begin
  Result := TBiforcazione.Create(Self);
end;

function TBifurcation.GetLyapunov: TFunzione;
begin
  Result := TLyapunov.Create(Self);
end;

function TBifurcation.GetIterazione(L: Extended): TFunzione;
begin
  Result := TIterazione.Create(Self, L);
end;

procedure TBifurcation.SetSeed(Value: Extended);
begin
  Mutex.BeginWrite;
  FSeed := Value;
  Mutex.EndWrite;
end;

end.
