unit RPNCompiler;

interface

uses
  SysUtils, Classes, ASIConst;

type
  TEvaluator = function (Variabili: PExtendedArray0; Costanti: PExtendedArray0): Extended; pascal;

  TValidSet = set of Byte;

  PCurrentPila = ^TCurrentPila;
  TCurrentPila = record
    AC: Integer;
    PS: TList;
    PO: TList;
  end;

  PAtom = ^TAtom;
  TAtom = record
    Value: string;
    Float: Extended;
    Code: Integer;
    Tipo: Integer;
  end;

type
  TRPNCompiler = class(TComponent)
  private
    { Private declarations }
    Costanti: PExtendedArray0;
    ConstCount: Integer;
    ArgCount: Integer;
    Buffer: array[0..1000] of Byte;
    Evaluator: TEvaluator;
    FText: string;
    Pile: TList;
    ErrPos: Integer;
    StackCount: Integer;
    MaxStack: Integer;
    MS: TStream;
    TmpAC: Integer;

    procedure Clean;
    procedure CreaPila;
    procedure SetText(Value: string);
    procedure ProcessFunction(Value: string);
    procedure ScriviAsm;
    function GetCurrentPila: PCurrentPila;
    procedure SearchValid(var Funz: string; var Cmd: TAtom; var Letti: Integer);
    function CopyAtom(Cmd: PAtom): PAtom;
    procedure CheckFPStack(A: Integer);

    procedure Error(Msg: string);
    procedure ParseError(Msg: string);

    procedure Comma(var Valid: TValidSet; Cmd: PAtom);
    procedure Data(var Valid: TValidSet; Cmd: PAtom);
    procedure ParChiusa(var Valid: TValidSet; Cmd: PAtom);
    procedure ParAperta(var Valid: TValidSet; Cmd: PAtom);
    procedure OperContr(var Valid: TValidSet; Cmd: PAtom);
    procedure OperFunc(var Valid: TValidSet; Cmd: PAtom);
    procedure Finito(var Valid: TValidSet; Cmd: PAtom);

    procedure AsmData(Cmd: PAtom);
    procedure AsmOperContr(Cmd: PAtom);
    procedure AsmOperFunc(Cmd: PAtom);
    procedure AsmX(Cmd: PAtom);
    procedure AsmConst(Cmd: PAtom);
    procedure AsmDown(Dati: array of Byte);
  protected
    { Protected declarations }
    MREW: TMultiReadExclusiveWriteSynchronizer;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Eval(Args: array of Extended): Extended; overload;
    function Eval(Args: PExtendedArray0): Extended; overload;
    function Eval(Args: PExtendedArray0; Argc: Integer): Extended; overload;
    procedure Dump(Linee: TStrings);
  published
    { Published declarations }
    property Text: string read FText write SetText;
  end;

type
  RPNException = class(Exception);

procedure Register;

implementation

const
  arpnError     =       0;
  arpnNumber    =       1;
  arpnLeftPar   =       2;
  arpn4Oper     =       3;
  arpnFunc      =       4;
  arpnEnd       =       5;
  arpnVar       =       6;
  arpnRightPar  =       7;
  arpnConst     =       8;
  arpnComma     =       9;

  afFSin        =       1;
  afFCos        =       2;
  afFAbs        =       3;
  afFChs        =       4;
  afFInt        =       5;
  afFSqrt       =       6;
  afFPatan	=	7;
  afFPtan	=	8;
  afLog2	=	9;
  afLn		=	10;
  afFrac	=	11;
  afExp		=	12;
  afPos		=	13;

  acPi          =       1;

type
  TFunzPredef = record
    Name: string;
    Tipo: Integer;
    Code: Integer;
  end;

const
  Funzioni: array[1..14] of TFunzPredef = (
    (Name: 'sin';       Tipo: arpnFunc;         Code: afFSin),
    (Name: 'cos';       Tipo: arpnFunc;         Code: afFCos),
    (Name: 'abs';       Tipo: arpnFunc;         Code: afFAbs),
    (Name: 'chs';       Tipo: arpnFunc;         Code: afFChs),
    (Name: 'int';       Tipo: arpnFunc;         Code: afFInt),
    (Name: 'sqrt';      Tipo: arpnFunc;         Code: afFSqrt),
    (Name: 'atan';      Tipo: arpnFunc;         Code: afFPatan),
    (Name: 'tan';       Tipo: arpnFunc;         Code: afFPtan),
    (Name: 'log';       Tipo: arpnFunc;         Code: afLog2),
    (Name: 'ln';        Tipo: arpnFunc;         Code: afLn),
    (Name: 'frac';      Tipo: arpnFunc;         Code: afFrac),
    (Name: 'exp';       Tipo: arpnFunc;         Code: afExp),
    (Name: 'pos';       Tipo: arpnFunc;         Code: afPos),

    (Name: 'pi';        Tipo: arpnConst;        Code: acPi)
  );

procedure Register;
begin
  RegisterComponents('AndSoft Inc.', [TRPNCompiler]);
end;

function Sample(Variabili: PExtendedArray0; Costanti: PExtendedArray0): Extended; pascal;
begin
  raise RPNException.Create('RPN Compiler not initialized!');
end;

function OPrec(O: Char): Integer;
begin
  case O of
    '+', '-': Result := -6;
    '*', '/': Result := -4;
    '%':      Result := -2;
  else
    Result := 0;
  end;
end;

function ParseFloat(S: string; var X: Extended): Integer;
var
  I: Integer;
begin
  for I := Length(S) downto 1 do begin
    try
      X := StrToFloat(Copy(S, 1, I));
      Result := I;
      BREAK;
    except
    end;
  end;
end;

constructor TRPNCompiler.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Evaluator := Sample;
  MREW := TMultiReadExclusiveWriteSynchronizer.Create;
end;

destructor TRPNCompiler.Destroy;
begin
  Clean;
  MREW.Free;
  inherited Destroy;
end;

procedure TRPNCompiler.Error(Msg: string);
begin
  Clean;
  raise RPNException.Create(Msg);
end;

procedure TRPNCompiler.ParseError(Msg: string);
begin
  Error(Msg + ' at position: ' + IntToStr(ErrPos));
end;

function TRPNCompiler.CopyAtom(Cmd: PAtom): PAtom;
begin
  New(Result);
  Result^ := Cmd^;
end;

procedure TRPNCompiler.CheckFPStack(A: Integer);
begin
  Inc(A, StackCount);
  if A > 7 then
    Error('FPU stack overflow! Function is too complex!');
  if A > MaxStack then MaxStack := A;
end;

procedure TRPNCompiler.SearchValid(var Funz: string; var Cmd: TAtom; var Letti: Integer);
var
  K, V, I, J: Integer;
  O: Boolean;
  Tmp: string;
begin
  with Cmd do begin
    if Funz = '' then begin
      Tipo := arpnEnd;
      Exit;
    end;
    case Funz[1] of
      '0'..'9':
        begin
          Letti := ParseFloat(Funz, Float);
          Tipo := arpnNumber;
          Inc(ConstCount);
        end;
      '(':
        begin
          Tipo := arpnLeftPar;
          Letti := 1;
        end;
      ')':
        begin
          Tipo := arpnRightPar;
          Letti := 1;
        end;
      '+', '-', '*', '/', '%':
        begin
          Tipo := arpn4Oper;
          Letti := 1;
        end;
      ';':
        begin
          Tipo := arpnComma;
          Letti := 1;
        end;
      'x':
        begin
          Tmp := Copy(Funz, 2, Length(Funz) - 1);
          Val(Tmp, V, I);
          if I = 0 then J := Length(Tmp) else J := I - 1;
          if J = 0 then
            ParseError('Error parsing variable!');
          Code := V;
          Tipo := arpnVar;
          Letti := J + 1;
        end;
      else begin
        K := 1;
        O := False;
        Tipo := arpnError;
        repeat
          if Copy(Funz, 1, Length(Funzioni[K].Name)) = Funzioni[K].Name then begin
            Tipo := Funzioni[K].Tipo;
            Code := Funzioni[K].Code;
            Letti := Length(Funzioni[K].Name);
            O := True;
          end;
          Inc(K)
        until (K = Length(Funzioni) + 1) or O;
        if Tipo = arpnError then
          ParseError('Unknown function!');
      end;
    end;
    Value := Copy(Funz, 1, Letti);
    Delete(Funz, 1, Letti);
  end;
end;

procedure TRPNCompiler.Data(var Valid: TValidSet; Cmd: PAtom);
var
  CP: PCurrentPila;
begin
  CP := GetCurrentPila;
  CP.PO.Add(CopyAtom(Cmd));
  Valid := [arpn4Oper, arpnEnd, arpnRightPar, arpnComma];
end;

procedure TRPNCompiler.ParChiusa(var Valid: TValidSet; Cmd: PAtom);
var
  CP, TP: PCurrentPila;
  Tmp: PAtom;
begin
  if Pile.Count = 1 then
    ParseError('Cannot close a parenthesis here!');
  CP := GetCurrentPila;
  if CP.AC <> 1 then
    ParseError('Too few arguments!');
  while CP.PS.Count <> 0 do begin
    Tmp := CP.PS.Last;
    CP.PS.Remove(Tmp);
    CP.PO.Add(Tmp);
  end;
  Pile.Remove(CP);
  TP := GetCurrentPila;
  while CP.PO.Count <> 0 do begin
    Tmp := CP.PO.First;
    CP.PO.Remove(Tmp);
    TP.PO.Add(Tmp);
  end;
  CP.PO.Free;
  CP.PS.Free;
  FreeMem(CP);
  Valid := [arpn4Oper, arpnEnd, arpnRightPar, arpnComma];
end;

procedure TRPNCompiler.ParAperta(var Valid: TValidSet; Cmd: PAtom);
var
  CP: PCurrentPila;
begin
  New(CP);
  CP.AC := TmpAC;
  TmpAC := 1;
  CP.PS := TList.Create;
  CP.PO := TList.Create;
  Pile.Add(CP);
  Valid := [arpnNumber, arpnLeftPar, arpnFunc, arpnVar, arpnConst];
end;

procedure TRPNCompiler.Comma(var Valid: TValidSet; Cmd: PAtom);
var
  CP: PCurrentPila;
  Tmp: PAtom;
begin
  if Pile.Count = 1 then
    ParseError('Cannot have a comma here!');
  CP := GetCurrentPila;
  if CP.AC = 1 then
    ParseError('Too much arguments!');
  Dec(CP.AC);
  while CP.PS.Count <> 0 do begin
    Tmp := CP.PS.Last;
    CP.PS.Remove(Tmp);
    CP.PO.Add(Tmp);
  end;
  Valid := [arpnNumber, arpnLeftPar, arpnFunc, arpnVar, arpnConst];
end;

procedure TRPNCompiler.OperContr(var Valid: TValidSet; Cmd: PAtom);
var
  CP: PCurrentPila;
begin
  CP := GetCurrentPila;
  while (CP.PS.Count <> 0) and (OPrec(Cmd.Value[1]) <= OPrec(PAtom(CP.PS.Last).Value[1])) do begin
    CP.PO.Add(CP.PS.Last);
    CP.PS.Delete(CP.PS.Count - 1);
  end;
  CP.PS.Add(CopyAtom(Cmd));
  Valid := [arpnNumber, arpnLeftPar, arpnVar, arpnFunc, arpnConst];
end;

procedure TRPNCompiler.OperFunc(var Valid: TValidSet; Cmd: PAtom);
var
  CP: PCurrentPila;
begin
  CP := GetCurrentPila;
  CP.PS.Add(CopyAtom(Cmd));
  TmpAC := 1;
  Valid := [arpnLeftPar];
end;

procedure TRPNCompiler.Finito(var Valid: TValidSet; Cmd: PAtom);
var
  CP: PCurrentPila;
  Tmp: PAtom;
begin
  if Pile.Count <> 1 then
    ParseError('Not enough parenthesis!');
  CP := GetCurrentPila;
  while CP.PS.Count <> 0 do begin
    Tmp := CP.PS.Last;
    CP.PS.Remove(Tmp);
    CP.PO.Add(Tmp);
  end;
end;

function TRPNCompiler.GetCurrentPila: PCurrentPila;
begin
  if Pile <> nil then
    Result := PCurrentPila(Pile.Last)
  else
    Result := nil;
end;

procedure TRPNCompiler.Clean;
var
  CP: PCurrentPila;
  Tmp: PAtom;
begin
  if Pile <> nil then begin
    while Pile.Count <> 0 do begin
      CP := PCurrentPila(Pile.First);
      while CP.PO.Count <> 0 do begin
        Tmp := CP.PO.First;
        FreeMem(Tmp);
        CP.PO.Remove(Tmp);
      end;
      CP.PO.Free;
      while CP.PS.Count <> 0 do begin
        Tmp := CP.PS.First;
        FreeMem(Tmp);
        CP.PS.Remove(Tmp);
      end;
      CP.PS.Free;
      FreeMem(CP);
      Pile.Remove(CP);
    end;
    Pile.Free;
    Pile := nil;
  end;
  ArgCount := 0;
  FText := '';
  FreeMem(Costanti);
  Costanti := nil;
  ConstCount := 0;
  ErrPos := 0;
  MaxStack := 0;
  TmpAC := 1;
  Evaluator := Sample;
end;

procedure TRPNCompiler.CreaPila;
var
  CP: PCurrentPila;
begin
  Pile := TList.Create;
  New(CP);
  CP.AC := 1;
  CP.PS := TList.Create;
  CP.PO := TList.Create;
  Pile.Add(CP);
end;

procedure TRPNCompiler.SetText(Value: string);
begin
  if Value <> FText then begin
    try
      MREW.BeginWrite;
      Clean;
      CreaPila;
      FText := Value;
      ProcessFunction(Value);
      ScriviAsm;
      Evaluator := TEvaluator(@Buffer);
    finally
      MREW.EndWrite;
    end;
  end;
end;

function TRPNCompiler.Eval(Args: array of Extended): Extended;
begin
  Result := Eval(@Args, Length(Args));
end;

function TRPNCompiler.Eval(Args: PExtendedArray0): Extended;
begin
  try
    Result := Eval(Args, ArgCount + 1);
  finally
  end;
end;

function TRPNCompiler.Eval(Args: PExtendedArray0; Argc: Integer): Extended;
begin
  try
    if Argc >= ArgCount + 1 then
      Result := Evaluator(Args, Costanti)
    else
      Error('Too few arguments!');
  finally
  end;
end;

procedure TRPNCompiler.ProcessFunction(Value: string);
var
  Valid: TValidSet;
  Temp: string;
  Cmd: TAtom;
  Letti: Integer;
begin
  Valid := [arpnNumber, arpnLeftPar, arpnVar, arpnConst, arpnFunc];
  Temp := FText;
  repeat
    SearchValid(Temp, Cmd, Letti);
    if not(Cmd.Tipo in Valid) then
      ParseError('Function is not legal here!');
    case Cmd.Tipo of
      arpnNumber, arpnVar, arpnConst:   Data(Valid, @Cmd);
      arpnLeftPar:                      ParAperta(Valid, @Cmd);
      arpn4Oper:                        OperContr(Valid, @Cmd);
      arpnFunc:                         OperFunc(Valid, @Cmd);
      arpnEnd:                          Finito(Valid, @Cmd);
      arpnRightPar:                     ParChiusa(Valid, @Cmd);
      arpnComma:                        Comma(Valid, @Cmd);
    else
      Error('Internal error!');
    end;
    Inc(ErrPos, Letti)
  until Cmd.Tipo = 5;
end;

procedure TRPNCompiler.Dump(Linee: TStrings);
var
  CP: PCurrentPila;
  I: Integer;
  Cmd: PAtom;
begin
  try
    MREW.BeginRead;
    CP := GetCurrentPila;
    if CP <> nil then
      for I := 0 to CP.PO.Count - 1 do begin
        Cmd := PAtom(CP.PO.Items[I]);
        Linee.Add(Cmd.Value);
      end
    else
      Error('No function defined!');
  finally
    MREW.EndRead;
  end;
end;

procedure TRPNCompiler.AsmDown(Dati: array of Byte);
begin
  MS.Write(Dati, Length(Dati));
end;

procedure TRPNCompiler.AsmData(Cmd: PAtom);
begin
  CheckFPStack(1);
  Costanti[ConstCount] := Cmd.Float;
  if ConstCount = 0 then
    AsmDown([$DB, $2A])                   (* fld tbyte ptr [edx]       *)
  else
    AsmDown([$DB, $6A, ConstCount * 10]); (* fld tbyte ptr [edx + $00] *)
  Inc(ConstCount);
  Inc(StackCount);
end;

procedure TRPNCompiler.AsmOperContr(Cmd: PAtom);
begin
  Dec(StackCount);
  case Cmd.Value[1] of
    '+': AsmDown([$DE, $C1]);
    '/': AsmDown([$DE, $F9]);
    '*': AsmDown([$DE, $C9]);
    '-': AsmDown([$DE, $E9]);
    '%': Asmdown([$D9, $C9,		        (* fxch   *)
                  $D9, $F5,		        (* fprem1 *)
    		  $DD, $D9]);		        (* fstp st(1) *)
  else
    Error('Internal error!');
  end;
end;

procedure TRPNCompiler.AsmOperFunc(Cmd: PAtom);
begin
  case Cmd.Code of
    afFSin:   AsmDown([$D9, $FE]);
    afFCos:   AsmDown([$D9, $FF]);
    afFInt:   AsmDown([$D9, $FC]);
    afFSqrt:  AsmDown([$D9, $FA]);
    afFChs:   AsmDown([$D9, $E0]);
    afFAbs:   AsmDown([$D9, $E1]);
    afFPatan: begin
    	      CheckFPStack(1);
              AsmDown([$D9, $E8,		(* fld1   *)
                       $D9, $F3]);		(* fpatan *)
              end;
    afFPtan:  begin
    	      CheckFPStack(1);
              AsmDown([$D9, $F2,		(* fptan  *)
                       $DD, $D8]);		(* fstp st(0) *)
              end;
    afLog2:   begin
    	      CheckFPStack(1);
              AsmDown([$D9, $E8,		(* fld1	  *)
                       $D9, $C9,		(* fxch	  *)
                       $D9, $F1]);		(* fyl2x  *)
              end;
    afLn:     begin
    	      CheckFPStack(1);
              AsmDown([$D9, $ED,		(* fldl2n *)
                       $D9, $C9,		(* fxch   *)
                       $D9, $F1]);		(* fyl2x  *)
              end;
    afFrac:   begin
    	      CheckFPStack(1);
              AsmDown([$D9, $C0,		(* fld st(0) *)
                       $D9, $FC,		(* frndint *)
                       $DE, $E9]);		(* fsubp   *)
              end;
    afExp:    begin
    	      CheckFPStack(2);
              AsmDown([$D9, $EA,		(* fldl2e  *)
                       $DE, $C9,		(* fmulp   *)
                       $D9, $C0,		(* fld st(0) *)
		       $D9, $FC,		(* frndint *)
		       $DC, $E9,		(* fsub st(1), st(0) *)
		       $D9, $C9,		(* fxch    *)
		       $D9, $F0,		(* f2xm1   *)
		       $D9, $E8,		(* fld1    *)
		       $DE, $C1,		(* faddp   *)
		       $D9, $FD,		(* fscale  *)
		       $DD, $D9]);		(* fstp st(1) *)
              end;
    afPos:    begin
    	      CheckFPStack(1);
              AsmDown([$D9, $E4,		(* ftst    *)
                       $DF, $E0,		(* fnstsw ax *)
                       $9E, 			(* sahf    *)
                       $D9, $EE,  		(* fldz    *)
                       $DB, $D1,		(* fcmovnbe st(0), st(1) *)
                       $DD, $D9]);		(* fstp st(1) *)
              end;
  else
    Error('Internal error!');
  end;
end;

procedure TRPNCompiler.AsmConst(Cmd: PAtom);
begin
  CheckFPStack(1);
  case Cmd.Code of
    acPi: AsmDown([$D9, $EB]);
  else
    Error('Internal error!');
  end;
  Inc(StackCount);
end;

procedure TRPNCompiler.AsmX(Cmd: PAtom);
begin
  CheckFPStack(1);
  if Cmd.Code > ArgCount then ArgCount := Cmd.Code;
  if Cmd.Code = 0 then
    AsmDown([$DB, $29])                 (* fld tbyte ptr [ecx]       *)
  else
    AsmDown([$DB, $69, Cmd.Code * 10]); (* fld tbyte ptr [ecx + $00] *)
  Inc(StackCount);
end;

procedure TRPNCompiler.ScriviAsm;
var
  Size: Integer;
  CP: PCurrentPila;
  Cmd: PAtom;
  I: Integer;
begin
  MS := TMemoryStream.Create;
  try
    AsmDown([$8B, $4C, $24, $08,           (* mov  ecx,  [esp + $08] *)
             $8B, $54, $24, $04,           (* mov  edx,  [esp + $04] *)
             $DB, $E3]); 		   (* fninit                 *)
    CP := GetCurrentPila;
    GetMem(Costanti, ConstCount * SizeOf(Extended));
    ConstCount := 0;
    StackCount := 0;
    MaxStack := 0;
    for I := 0 to CP.PO.Count - 1 do begin
(*    while CP.PO.Count <> 0 do begin
      Cmd := PAtom(CP.PO.First); *)
      Cmd := PAtom(CP.PO.Items[I]);
        case Cmd.Tipo of
          arpnNumber:     AsmData(Cmd);
          arpn4Oper:      AsmOperContr(Cmd);
          arpnFunc:       AsmOperFunc(Cmd);
          arpnVar:        AsmX(Cmd);
          arpnConst:      AsmConst(Cmd);
        else
          Error('Internal error!');
        end;
    end;
    AsmDown([$9B,                          (* wait                   *)
             $C2, $08, $00]);              (* ret  $0008             *)
    Size := MS.Seek(0, soFromEnd);
    MS.Seek(0, soFromBeginning);
    MS.Read(Buffer, Size);
    Evaluator := TEvaluator(@Buffer);
  finally
    MS.Free;
  end;
end;

end.


